# -*- coding: utf-8 -*-
"""
    msiem session
"""
import ast
import re
import base64
import requests
import urllib3
import json
import getpass
import ssl
from urllib.parse import urljoin, unquote
from abc import abstractmethod
from concurrent.futures import ThreadPoolExecutor
from abc import ABC
urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

from .params import PARAMS
from .constants import BASE_URL, BASE_URL_PRIV, ASYNC_MAX_WORKERS, GENERAL_POST_TIMEOUT
from .exceptions import ESMException
from .utils import tob64, getLogger
from .config import ESMConfig
        

class ESMSession():

    """
        # static variables
        _headers={'Content-Type': 'application/json'}
        _executor = ThreadPoolExecutor(max_workers=ASYNC_MAX_WORKERS)

        _config=None
    

    #Singleton : unique instance
    _instance = None

    def __new__(cls, *args, **kwargs):
        if ESMSession._instance is None :
            self._logger.debug('Creating a new instance of ESMSession')
            ESMSession._instance = ABC.__new__(cls, *args, **kwargs)

        return ESMSession._instance
        """

    _initiated = False
    _shared_state = {}

    def __init__(self, conf_path=None, **config):
        self.__dict__ = self._shared_state
        if not self._initiated :
            #Config parsing
            self._config = ESMConfig(path=conf_path, **config)
            self._logger = getLogger(v=self._config.getboolean('general', 'verbose'))
            self._executor = ThreadPoolExecutor(max_workers=ASYNC_MAX_WORKERS)
            self._headers={'Content-Type': 'application/json'}
            self._logged=False
            self._initiated = True
            
    def __str__(self):
        return repr(self)

    def _login(self):
        
        userb64 = tob64(self._config.get('esm', 'user'))
        passb64 = self._config.get('esm', 'passwd')
        
        try :
            resp = self.esmRequest('login', username=userb64, password=passb64, raw=True, secure=True)
            self._headers['Cookie'] = resp.headers.get('Set-Cookie')
            self._headers['X-Xsrf-Token'] = resp.headers.get('Xsrf-Token')
        
        except Exception :
            raise ESMException("Login failed")

        else :
            return True

    """
    def logout(self):
        self.esmRequest('logout')
        
    """

    def esmRequest(self, method, callback=None, raw=False, secure=False, **params):

        self._logger.debug("Calling esmRequest with params :"+str(params))

        method, data = PARAMS.get(method)

        if data :
            data =  data % params
            data = ast.literal_eval(''.join(data.split()))
        
        if method :
            method = method % params

        post=dict()

        if not self._logged and method != 'login':
            self._logged=self._login()

        try :
            post = self._post(method, data, callback, raw, secure)
            return post

        except Exception as err:
            self._logger.error(str(err))
            raise err

    def _post(self, method=None, data=None, callback=None, raw=False, secure=False):
        """
        Helper method
        If method is all upper cases, a private API is done.
        Private API is under /ess/ and public api is under /rs/esm
        """

        url=str()
        privateApiCall=False
        result=None

        #Handling private API calls formatting
        if method == method.upper():
            privateApiCall=True
            url = BASE_URL_PRIV
            if data :
                data = self._format_params(method, **data)
        
        #Normal API calls
        else:
            url = BASE_URL
            if data:
                data = json.dumps(data)

        self._logger.debug('POSTING : ' + method + ('\nDATA'+(data if data is not None else '') if not secure else ''))
        #self._logger.debug(self.__dict__)

        try :
            result = requests.post(
                urljoin(url.format(self._config.get('esm', 'host')), method),
                data=data, 
                headers=self._headers,
                verify=self._config.getboolean('general','ssl_verify'),
                timeout=GENERAL_POST_TIMEOUT
            )

            if not secure :
                pass
                #self._logger.debug('RESULT : '+result.text)

            if raw :
                return result

            else:
                try:
                    result.raise_for_status()

                except requests.HTTPError as e :
                    self._logger.error(str(e)+' '+str(result.text))

                else: #
                    try:
                        result = result.json()
                        result = result.get('return')

                    except json.decoder.JSONDecodeError:
                        result = result.text

                    if privateApiCall :
                        result = self._format_priv_resp(result)

                    if callback:
                        result = callback(result)

                    return result

        except ConnectionError as e:
            self._logger.critical(e)
            raise
        except requests.exceptions.Timeout as e:
            self._logger.error(e)
            pass
        except requests.exceptions.TooManyRedirects as e :
            self._logger.error(e)
            pass
        except Exception as e:
            self._logger.error(e)
        
    @staticmethod
    def _format_params(cmd, **params):
        """
        Format private API call.
        From mfe_saw project at https://github.com/andywalden/mfe_saw
        """
        params = {k: v for k, v in params.items() if v is not None}
        params = '%14'.join([k + '%13' + v + '%13' for (k, v) in params.items()])
        
        if params:
            params = 'Request=API%13' + cmd + '%13%14' + params + '%14'
        else:
            params = 'Request=API%13' + cmd + '%13%14'
        return params

    @staticmethod
    def _format_priv_resp(resp):
        """
        Format response from private API
        From mfe_saw project at https://github.com/andywalden/mfe_saw
        """
        resp = re.search('Response=(.*)', resp).group(1)
        resp = resp.replace('%14', ' ')
        pairs = resp.split()
        formatted = {}
        for pair in pairs:
            pair = pair.replace('%13', ' ')
            pair = pair.split()
            key = pair[0]
            if key == 'ITEMS':
                value = pair[-1]
            else:
                value = unquote(pair[-1])
            formatted[key] = value
        return formatted

class ESMObject(ABC):

    def __init__(self):
        self._session = ESMSession()

    def esmRequest(self, *args, **kwargs):
        return self._session.esmRequest(*args, **kwargs)
    




# -*- coding: utf-8 -*-
"""
    msiem config
"""

from configparser import ConfigParser, NoSectionError, MissingSectionHeaderError
import os
import sys
from getpass import getpass
from .utils import tob64
from .constants import DEFAULT_CONFIG_FILE, CONFIG_FILE_NAME



class ESMConfig(ConfigParser):
    """
    Find the config settings which include:
    
    [general]
    output=<text/json>
    verbose=<yes/no>
    ssl_verify=<yes/no>
    log_file=<path>

    [esm]
    host=<fqdn>
    user=<username>
    passwd=<b64password>
    
    
    ESMConfig(esm={'host':'10.2.2.1', 'user':'tristan'})
    """

    def __init__(self, path=None, **kwconfig):
        """
        Initialize the Config instance.
        """
        super().__init__()

        if not path :
            self._path = self._find_ini_location()

        else : 
            self._path = path

        try :
            f=self.read(self._path)
            if len(f) == 0:
                raise FileNotFoundError

        except :
            print("Config file inexistant or currupted, applying defaults")

            if not os.path.exists(os.path.dirname(self._path)):
                os.makedirs(os.path.dirname(self._path))

            with open(self._path, 'w') as conf :
                conf.write(DEFAULT_CONFIG_FILE)

            self.read(self._path)
            self.write()

        if kwconfig :
            for section in kwconfig :
                try :
                    if self.has_section(section):
                        for key in kwconfig[section] :
                            #log.info("Setting ["+section+"] "+key+" = "+kwconfig[section][key])
                            self.set(section, key, kwconfig[section][key])
                    else:
                        pass
                        #log.warning("Skipping unknown config section "+section)

                except Exception as err:
                    raise Exception("An error occured when reading the manual config dict. \n"+str(err))
        
    def write(self, path=None):
        if not path :
            path=self._path
            if not path :
                path = self._find_ini_location()
        
        if path :
            try :
                with open(path, 'w') as conf:
                    super().write(conf)

                self._path=path
                print("Config file wrote at "+path)

            except Exception as err :
                raise Exception("An error occured when writing the config file. \n"+str(err))
        else :
            raise Exception("Couldn't write config file, no specified path.")

    @staticmethod
    def _find_ini_location():
        """
        Attempt to locate the conf.ini file 
        """
        conf_path=None

        if 'APPDATA' in os.environ:
            conf_path = os.environ['APPDATA']

        elif 'XDG_CONFIG_HOME' in os.environ:  
            conf_path = os.environ['XDG_CONFIG_HOME']

        elif 'HOME' in os.environ:  
            conf_path = os.path.join(os.environ['HOME'])
    
       
        if conf_path  :
            conf_path=(os.path.join(conf_path, CONFIG_FILE_NAME))
            #log.info("Assuming config file at {}".format(conf_path))
        else :
            raise SystemError("Couldn't find either APPDATA, XDG_CONFIG_HOME or HOME environemnt variable to locate the conf.ini file.")

        return(conf_path)

    def setAuth(self):
        self.set('esm', 'host', input('Please enter the ESM FQDN: '))
        self.set('esm', 'user', input('Please enter your username: '))
        self.set('esm', 'passwd', tob64(getpass()))

    def set(self, section, option, value=None):
        if value == '' :
            pass
        else:
            super().set(section, option, value)

    def iset(self, msg, section, option):
        value = self.get(section, option)
        newvalue=''
        if option is 'passwd' :
            newvalue = tob64(getpass(msg+ '. Press <Enter> to skip: '))
        else:
            newvalue = input(msg+ '. Press <Enter> to keep '+ (value if (str(value) is not '') else 'empty') + ': ')
        self.set(section, option, newvalue)
        
    def interactiveSet(self):
        self.iset('Enter ESM hostname', 'esm', 'host')
        self.iset('Enter ESM username', 'esm', 'user')
        self.iset('Enter ESM password', 'esm', 'passwd')
        self.write()

    def show(self):
        print('Configuration file : '+self._path+'\n'+open(self._path, 'r').read())


# -*- coding: utf-8 -*-
"""
    msiem command
"""

import argparse
from .config import ESMConfig
from .session import ESMSession
from .query import AlarmQuery, Alarm, EventQuery, Event
from .constants import POSSIBLE_TIME_RANGE, ALARM_FILTER_FIELDS, ALARM_EVENT_FILTER_FIELDS

lol="""
What it could look like :

msiem 
    --help #Exits after
    --version #Exists after
    --batch

msiem config
    --help
    --list/-l
    --set auth
    --set <section.key>=<value>

msiem esm 
    --help/-h
    --time/-t
    --version/-v
    --tz/--timezones
    --disks/-d
    --callhome
    --ram
    --status
    --backup-status
    --buildstamp
    --recs
    --rule

    #Todo
    --nslookup
    --fields -> display list of possible fields with qryGetSelectFields
    --filters -> display list of possible filters with qryGetFilterFields

    #FullBackUp
    --backup esm_full_backup
    https://github.com/andywalden/esm_full_backup/

msiem ds
    --help/-h
    --add/-a interactive si manque arg
        name=DC01_DNS 
        ip=10.10.1.34 
        rec_ip=172.16.15.10
        type=linux
        Will add a new datasource and also call --info
        Ask for confirmation unless --force

    --del/-d <id or name> if resolves to only one device
        Ask for confirmation unless --force

    --search/-s <pattern>
    --searchgroup/-g
    --info/-i <id or name>
    --times/-t
    (--refresh)
    --recs --> ?
    --ipvalid <>
    --addclient <ds> <> ...

    --checkinactive
    Queries a McAfee ESM for inactive data sources. from https://github.com/andywalden/esmcheckds2

msiem alarms

    --search/-s [default]

    --ack/-a
        Ask for confirmation unless --force

    --del/-d
        Ask for confirmation unless --force

    --unack/-u

    
   
    --time_range LAST_24_HOURS
    --start_time 2019-03-25
    --end_time 2019-03-25
    --status [acked/unacked/all]

    --time 2019-03-25T14:00
    --window 1h

    --filter/-f    
                '[event.]srcip 10.2.2.2'
                '[event.]dstip 10.2.2.1'
                '[event.]msg=http-vulnerability'
                '[event.]count=http-vulnerability'
                '[event.]protocol=http-vulnerability'
                'event.time=http-vulnerability'
                '[event.]subtype=http-vulnerability'

                summary=<>
                severity=90
                ackuser=tristan
                assignee=tristan
                name=Critical
                acktime

                

                #No filtering on DetailedAlarm fields value !
                #Only a couple basic event related fields are available for filtering
                #Filters are ANDed together

    --info/-i <id>
        --> show the conditionnal xml tree and alarm meta data

    --force


msiem events
    --search

    --filter
            -f sourceip=10.2.2.2
            -f destinationip=10.2.2.1
            -f message=http-vulnerability
            -f 'name=Bob le bricoleur'

            -f time=2019-03-25T14:00
            -f window=1h
            -f timerange=LAST_24_HOURS
            -f begins=2019-03-25
            -f ends=2019-03-25

            #and others under event. #See documentation
            #Filters are ANDed together

    --fields
        #Fields to add to the defaults settings

    --order <order>

msiem watchlist
    Inteface to whatchlist operations

msiem service
    Opens a service now mcfee ticket.
    Callable by the Execute function of the ESM.
    https://github.com/andywalden/mfe2snow

msiem syslog
    Generate syslog message on the SIEM

msiem case
    --close https://github.com/andywalden/esm_close_cases
    --open <alarm>
    --msg <message>

msiem elm 
    Export ELM files back to original format
    https://github.com/andywalden/elmex/blob/master/elmex.sh
    
"""


def parseArgs():
    parser = argparse.ArgumentParser(description='McAfee SIEM Command Line Interface and Python API',
                usage='Use "msiem --help" for more information',
                formatter_class=argparse.RawTextHelpFormatter)

    parser.add_argument('--version', help="Show version",action="store_true")
    parser.add_argument('-v', '--verbose', help="Increase output verbosity",action="store_true")

    commands = parser.add_subparsers()

    config = commands.add_parser('config')
    config.set_defaults(func=config)
    config.add_argument('--list', help="List configuration fields", action="store_true")
    config.add_argument('--set', help="Will inveractively prompt for configuration settings", action="store_true")

    alarm = commands.add_parser('alarms')
    alarm.set_defaults(func=alarms)
    alarm.add_argument('--ack', help="Acknowledge the alarms", action="store_true")
    alarm.add_argument('--unack', help="Unacknowledge the alarms", action="store_true")
    alarm.add_argument('--delete', help="Delete the alarms", action="store_true")
    
    alarm.add_argument('--time_range','-t', metavar='time_range', help='Timerange in'+str(POSSIBLE_TIME_RANGE))
    alarm.add_argument('--start_time','--t1', metavar='time', help='Start trigger date')
    alarm.add_argument('--end_time','--t2', metavar='time', help='End trigger date')

    alarm.add_argument('--status', metavar='status', help='Status of the alarm [ack|unack|all]')

    alarm.add_argument('--filters', '-f', metavar="'<key>=<match>'", nargs='+', type=str, help="List of filters")

    alarm.add_argument('--summary', metavar="sumary", help="Alarm summary filter")
    alarm.add_argument('--assignee', metavar="assignee", help="Alarm assignee filter")
    alarm.add_argument('--severity', metavar="severity", help="Alarm severity filter")
    alarm.add_argument('--trigdate', metavar="trigdate", help="Alarm trigdate filter")
    alarm.add_argument('--ackdate', metavar="ackdate", help="Alarm ackdate filter")
    alarm.add_argument('--ackuser', metavar="ackuser", help="Alarm ackuser filter")
    alarm.add_argument('--name', metavar="name", help="Alarm name filter")

    alarm.add_argument('--msg', metavar="msg", help="Event msg filter")
    alarm.add_argument('--count', metavar="count", help="Event count filter")
    alarm.add_argument('--srcip', metavar="srcip", help="Event srcip filter")
    alarm.add_argument('--dstip', metavar="dstip", help="Event dstip filter")
    alarm.add_argument('--protocol', metavar="protocol", help="Event protocol filter")
    alarm.add_argument('--date', metavar="date", help="Event date filter")
    alarm.add_argument('--subtype', metavar="subtype", help="Event subtype filter")

    return (parser.parse_args())

def config(args):

    if args.set :
        ESMConfig().interactiveSet()

    if args.list :
        ESMConfig().show()

def alarms(args):
    
    vargs=vars(args)

    #Filter out the filters keys from list of parameters based on their names
    filters = list()
    for key in vargs :
        if vargs[key] is not None:
            for synonims in ALARM_FILTER_FIELDS+ALARM_EVENT_FILTER_FIELDS:
                if key in synonims :
                    filters.append((key, vargs[key]))
    
    #Merging values from --filters option
    if args.filters is not None :
        filters+=args.filters

    print(filters)

    alarms=AlarmQuery(
        time_range=args.time_range,
        start_time=args.start_time,
        end_time=args.end_time,
        status=args.status,
        filters=filters
    ).execute()

    if len(alarms) >0:

        alarms.show()

        if args.ack and 'y' in input('Are you sure you want to acknowledge those alarms ? [y/n]'):
            alarms.acknowledge()

        if args.unack and 'y' in input('Are you sure you want to unacknowledge those alarms ? [y/n]'):
            alarms.unacknowledge()

def main():
    args = parseArgs()

    if hasattr(args, 'func') :
        args.func(args)
    

if __name__ == "__main__":
    main()





# -*- coding: utf-8 -*-
"""
    msiem query
"""
import json
import time
import concurrent.futures
from tqdm import tqdm
from abc import abstractmethod, abstractproperty
from prettytable import PrettyTable
from .session import ESMObject, ESMSession
from .exceptions import ESMException
from .utils import getTimes, regexMatch
from .constants import (POSSIBLE_TIME_RANGE,
    POSSIBLE_FIELD_TYPES, 
    POSSIBLE_OPERATORS, 
    POSSIBLE_VALUE_TYPES,
    POSSIBLE_ALARM_STATUS,
    POSSBILE_ROW_ORDER,
    DEFAULTS_EVENT_FIELDS,
    ALARM_FILTER_FIELDS,
    ALARM_EVENT_FILTER_FIELDS,
    ALARM_DEFAULT_FIELDS)

class QueryBase(ESMObject):
    def __init__(self, time_range=None, start_time=None, end_time=None):
        super().__init__()

        self._session._logger.debug("Creating query with times : "+str(locals()))

        #Declaring attributes
        self._time_range=str()
        self._start_time=str()
        self._end_time=str()

        timeIssue=True
        
        #init attributes
        if time_range is not None :
            self.time_range=time_range
            timeIssue=False

        if start_time is not None and end_time is not None :
            self.time_range='CUSTOM'
            self._start_time=start_time
            self._end_time=end_time
            timeIssue=False

        if timeIssue :
            raise ESMException("The query must have valid time specifications. Please refer to documentation.")

    def __str__(self):
        return str(self.__dict__)

    @property
    def time_range(self):
        return self._time_range

    @property
    def start_time(self):
        return self._start_time

    @property
    def end_time(self):
        return self._end_time
    

    @time_range.setter
    def time_range(self, time_range):
        """
        Set the time range of the query to the specified string value. Default : LAST_3_DAYS
        """

        if time_range in POSSIBLE_TIME_RANGE :
            self._time_range=time_range
            
        else:
            raise ESMException("The time range must be in "+str(POSSIBLE_TIME_RANGE))

    
    @start_time.setter
    def start_time(self, start_time):
        """
        Set the time start of the query. Time range must be CUSTOM for this to work.
        """
        
        if start_time is not None :
            if self._time_range == 'CUSTOM':
                self._start_time=start_time
                
            else:
                raise ESMException("The time range must be 'CUSTOM' if you want to specify a custom start time")
        else:
            self._start_time=None
                
    
    @end_time.setter
    def end_time(self, end_time):
        """
        Set the time end of the query. Time range must be CUSTOM for this to work.
        """
       
        if end_time is not None :
            if self._time_range == 'CUSTOM':
                self._end_time=end_time
                
            else:
                raise ESMException("The time range must be 'CUSTOM' if you want to specify a custom end time")
        else :
            self._end_time=None
                
    @abstractproperty
    def filters(self):
        raise ESMException("Not implemented in the base query")

    @filters.setter
    def filters(self, filters):
        
        if isinstance(filters, list):
            for f in filters :
                self.add_filter(f)

        elif isinstance(filters, tuple):
            self.add_filter(filters)

        elif filters is None :
            pass
        
        else :
            raise ESMException("Illegal type for the filter object, it must be a list of a tuple.")
       
    @abstractmethod
    def add_filter(self, filter):
        raise ESMException("Not implemented in the base query")

    @abstractmethod
    def execute(self):
        raise ESMException("Not implemented in the base query")

class TestingQuery(QueryBase):

    @property
    def filters(self):
        raise ESMException("Not implemented in the test query")
    
    def add_filter(self, filter):
        raise ESMException("Not implemented in the test query")
    
    def execute(self):
        raise ESMException("Not implemented in the test query")

class AlarmQuery(QueryBase):

    def __init__(self, status='', page_size=5000, page_number=1, filters=None, **args):

        """
        filters : [(field, [values]), (field, [values])]
        field can be an EsmTriggeredAlarm or an EsmTriggeredAlarmEvent field
        """

        super().__init__(**args)

        #Declaring attributes
        self._status=str()
        self._page_size=int()
        self._page_number=int()

        self._alarm_filters = list(tuple())
        self._event_filters = list(tuple())

        #Setting attributes
        self.status=status
        self.page_size=page_size
        self.page_number=page_number

        #uses the parent filter setter
        super(self.__class__, self.__class__).filters.__set__(self, filters)

    @property
    def status(self):
        return self._status

    @property
    def page_size(self):
        return self._page_size

    @property
    def page_number(self):
        return self._page_number

    @property
    def filters(self):
        return self._alarm_filters + self._event_filters

    @status.setter
    def status(self, status):
        """
        Set the status filter of the alarm query. 'acknowledged', 'unacknowledged', '' or null -> all (default is '')
        """

        if type(status) is str : 
            if status.lower() in POSSIBLE_ALARM_STATUS :
                self._status=status
                
            else:
                raise ESMException("Illegal value of status filter. The status must be in "+str(POSSIBLE_ALARM_STATUS))

    @page_size.setter
    def page_size(self, page_size):
        """
        Set the page size, the number of alarms to return per page (default is 5000, max is 5000).
        """
        try:
            if page_size > 0 and page_size <= 5000 :
                self._page_size=page_size
                
            else:
                raise ESMException("The page size must be in 0-5000")

        except:raise

    @page_number.setter
    def page_number(self, page_number):
        """
        Set the page number. Which page of alarms we want to return (default is 1)
        """
        self._page_number=page_number

    def add_filter(self, afilter):
        """
            Make sure the filters format is tuple(field, list(values in string))
            Takes also care of the differents synonims fields can have
        """

        if isinstance(afilter,str):
            afilter = afilter.split('=',1)
        try:
            values = afilter[1] if isinstance(afilter[1], list) else [afilter[1]]
            values = [str(v) for v in values]
            added=False

            for synonims in ALARM_FILTER_FIELDS :
                if afilter[0] in synonims :
                    self._alarm_filters.append((synonims[0], values))
                    added=True

            for synonims in ALARM_EVENT_FILTER_FIELDS :
                if afilter[0] in synonims :
                    self._event_filters.append((synonims[0], values))
                    added=True

        except IndexError:
            added = False

        if not added :
            raise ESMException("Illegal filter field value : "+afilter[0]+". The filter field must be in :"+str(ALARM_FILTER_FIELDS + ALARM_EVENT_FILTER_FIELDS))

    def execute(self):
        """"
        Execute the query.
        Returns a list of Alarms.
        """

        resp=None

        self._session._logger.debug("Query state at the moment of execution : "+str(self.__dict__))

        if self.time_range == 'CUSTOM' :
            resp=self.esmRequest(
                'get_alarms_custom_time',
                time_range=self.time_range,
                start_time=self.start_time,
                end_time=self.end_time,
                status=self.status,
                page_size=self.page_size,
                page_number=self.page_number
                )

        else :
            resp=self.esmRequest(
                'get_alarms',
                time_range=self.time_range,
                status=self.status,
                page_size=self.page_size,
                page_number=self.page_number
                )

        alarms=list()
        for alarm_data in resp :
            alarm = Alarm(**alarm_data)
            alarms.append(alarm)
        
        if len(alarms) == 5000:
            self._session._logger.warning("The maximum amount of alarms was retreived from the SIEM, some alarms are ignored refine your time range or status to avoid this.")

        return AlarmCollection(self._filter(alarms))

    def _alarm_match(self, alarm):
        match=True
        for alarm_filter in self._alarm_filters :
            match=False
            value = str(alarm.__dict__[alarm_filter[0]]) #Can only match strings
            for filter_value in alarm_filter[1]:
                if regexMatch(filter_value.lower(), value.lower()):
                    match=True
                    break
            if not match :
                break
        return match
        
    def _event_match(self, alarm):
        match=True
        for event_filter in self._event_filters :
            match=False
            values = [str(event[event_filter[0]]) for event in alarm.events] #Can only match strings
            for filter_value in event_filter[1]:
                if any(regexMatch(filter_value.lower(), value.lower()) for value in values):
                    match=True
                    break
            if not match :
                break
        return match
        
    @staticmethod
    def _detailled_alarm(alarm):
        return alarm.detailed

    def _filter(self, alarms, alarmonly=False):
        
        alarms = [a for a in alarms if self._alarm_match(a)]
        
        detailed=None

        if not alarmonly :
            self._session._logger.info("Getting alarms infos... Please be patient.")
            detailed = list(tqdm(self._session._executor.map(self._detailled_alarm, alarms), total=len(alarms)))
            alarms = [a for a in detailed if self._event_match(a)]

        self._session._logger.info(str(len(alarms)) + " alarms matching your filter(s)")
        return alarms

class Alarm(ESMObject):
    """
    Alarm object. 
    Gives possibility to access alarm fields and related events.
    Delete, Ack, or Unack alarm.
    """

    def __init__(self, **args):
        
        self.id = {"value" : 0}
        self.summary = ''
        self.assignee = ''
        self.severity = 0
        self.triggeredDate = ''
        self.acknowledgedDate = ''
        self.acknowledgedUsername = ''
        self.alarmName = ''
        self.conditionType = 0

        self.__dict__.update(args)

        self._detailed = None

        super().__init__()

    def _hasID(self):
        try :
            if self.id['value'] == 0 :
                return False
            else :
                return True
        
        except KeyError :
            return False


    @property
    def detailed(self):
        if self._hasID() :
            if self._detailed is None :
                self._detailed = DetailedAlarm(self)
            return self._detailed
        else :
            raise ESMException("Looks like this alarm doesn't have a valid ID. Cannot get the details.")

    @property
    def events(self):
        return self.detailed.events

    @property
    def status(self):
        return('acknowledged' if (
            (self.acknowledgedDate is None or len(self.acknowledgedDate)>0) 
            and 
            (self.acknowledgedUsername is None or len(self.acknowledgedUsername)>0)
            )
            else 'unacknowledged')

    def delele(self):
        if self._hasID() :
            self.esmRequest('delete_alarms', ids=[self.id['value']])
        else :
            raise ESMException("Looks like this alarm doesn't have a valid ID. Cannot delete.")
        return
    
    def acknowledge(self):
        if self._hasID() :
            self.esmRequest('ack_alarms', ids=[self.id['value']])
        else :
            raise ESMException("Looks like this alarm doesn't have a valid ID. Cannot acknowledge.")
        return

    def unacknowledge(self):
        if self._hasID() :
            self.esmRequest('unack_alarms', ids=[self.id['value']])
        else :
            raise ESMException("Looks like this alarm doesn't have a valid ID. Cannot unacknowledge.")
        return  

class AlarmCollection(list, ESMObject):
    """
        Double inheritance doesn't seem to work as expected, need to call ESMSession()
    """
    def __init__(self, alarms):
        super().__init__()
        self+=alarms

    def _ack(self, alarm):
        return alarm.acknowledge()

    def _unack(self, alarm):
        return alarm.unacknowledge()

    def acknowledge(self):
        ESMSession()._logger.info("Ackowledging alarms...")
        ESMSession()._executor.map(self._ack, self)

    def unacknowledge(self):
        ESMSession()._logger.info("Unackowledging alarms...")
        ESMSession()._executor.map(self._unack, self)

    def show(self, additionnalFields=[], sortBy=None):
        table = PrettyTable()
        table.field_names=[f[0] for f in ALARM_FILTER_FIELDS]+['status']+[f[0] for f in ALARM_EVENT_FILTER_FIELDS]
        for a in self :
            table.add_row(
                [a.__dict__[f[0]] for f in ALARM_FILTER_FIELDS]+[a.status]+
                [ (a.events[0][f[0]] if len(a.events)>0 else 'No event') for f in ALARM_EVENT_FILTER_FIELDS])

        print(table.get_string(fields=ALARM_DEFAULT_FIELDS+additionnalFields, sortby=sortBy))

    def json(self):
        return None

class DetailedAlarm(Alarm):
    """
    Alarm details object. Based on EsmTriggeredAlarmDetail
    """
    def __init__(self, alarm):

        if not alarm._hasID() :
            raise ESMException("Looks like this alarm doesn't have a valid ID. Cannot init DetailledAlarm.")
        
        super().__init__(id=alarm.id)

        """
        self.id = {"value" : 0} #duplicate
        self.summary = '' #duplicate
        self.assignee = '' #duplicate
        self.severity = 0 #duplicate
        self.triggeredDate  = '' #duplicate
        self.acknowledgedDate  = ''#duplicate
        self.acknowledgedUsername  = ''#duplicate
        self.alarmName  = ''#duplicate
        self.conditionType  = 0 #duplicate

        self.filters  = ''
        self.queryId = 0
        self.alretRateMin = 0
        self.alertRateCount = 0
        self.percentAbove = 0
        self.percentBelow = 0
        self.offsetMinutes = 0
        self.timeFilter  = ''
        self.maximumConditionTriggerFrequency = 0
        self.useWatchlist = ''
        self.matchField = ''
        self.matchValue = ''
        self.healthMonStatus = ''
        self.assigneeId = 0
        self.escalatedDate = ''
        self.caseId = 0
        self.caseName = ''
        self.iocName = 0
        self.iocId = 0
        self.description = ''
        self.actions = ''
        """

        self._events=list()

        resp = self.esmRequest('get_alarm_details', id=alarm.id)

        self.__dict__.update(resp)

        self.id = alarm.id

        self._events = resp['events']

    @property
    def detailed(self):
        return self

    @property
    def events(self):
        return self._events

class Event(ESMObject):
    """ Based on EsmTriggeredAlarmEvent """
    def __init__(self, **args):
        self.fields=dict().update(**args)

class EventQuery(QueryBase):
    """
    Interface to qryExecuteDetail?type=EVENT method.
    """

    #Declaring static value
    _possible_fields = []

    def __init__(self, fields=None, filters=None, 
        limit=5000, offset=0, order=None, compute_time_range=True, **args):
        """
       
        fields list of string, need to convert that into list of {
            "name": "(name)",
            "typeBits": 0,
            "id": "(id)"
            }

        order tuple (direction, field) to list of {
            "direction": "ASCENDING or DESCENDING",
            "field": {
                "name": ""
                }
            }

        filters 
            1 / list of {
                (field , [values]) #easy mode
            }

            2 / list of QueryFilter
        """

        self._compute_time_range = compute_time_range
        
        super().__init__(**args)

        #Constants
        self._type='EVENT'
        self._groupType='NO_GROUP'

        """ Not checking dynamically the velidity of the fields cause makes too much of unecessary requests
        #Singleton attribute mapping
        self._possible_fields = EventQuery._possible_fields
        if self._possible_fields is None :
            self._possible_fields = self.esmRequest('get_possible_fields', type=self._type, groupType=self._groupType)
            """

        #Declaring attributes
        self._filters=list()
        self._fields=set()

        self._limit=int()
        self._offset=int()

        self._reverse=bool()

        self._order=[{
            "direction": None,
            "field": {
                "name": None
                }
            }]

        self.fields = DEFAULTS_EVENT_FIELDS
        self.fields = fields

        self.filters = filters

        self.order = order
        
        self._limit = limit
        self._offset = offset

        self._query = None

        self._executed = False

    @property
    def fields(self):
        return([{'name':value} for value in list(self._fields)])

    @property
    def filters(self):
        return([f.configDict() for f in self._filters])

    @property
    def limit(self):
        return(self._limit)

    @property
    def offset(self):
        return(self._offset)

    @property
    def reverse(self):
        return(self._reverse)

    @property
    def order(self) :       
        return(self._order)

    @property
    def time_range(self):
        return(super().time_range)

    @order.setter
    def order(self, order):
        #tuple (direction, field)
        #TODO check valid value
        if type(order) is tuple :
            self._order[0]['direction']=order[0]
            self._order[0]['field']['name']=order[1]

    @filters.setter
    def filters(self, filters):
        """
            Uses BaseQuery.filter base implementation but adds a default filter when None
        """
        if not filters :
            self.filters = [FieldFilter(name='AvgSeverity', operator='GREATER_THAN' , values=[1])]
        else :
            super(self.__class__, self.__class__).filters.__set__(self, filters)
            #https://bugs.python.org/issue14965
            

    def add_filter(self, fil):
        if type(fil) is tuple :
            self._filters.append(FieldFilter(*fil))

        elif isinstance(fil, QueryFilter) :
            self._filters.append(fil)
        
        else :
            raise ESMException("Sorry the filters must be either a tuple(fiels, [values]) or a QueryFilter sub class.")

    @fields.setter
    def fields(self, fields):
        if fields :
            for f in fields :
                if f :
                    self.add_field(f) 

    def add_field(self, field):
        try:
            if True : # Not checking dynamically the velidity of the fields cause makes too much of unecessary requests any(f.get('name', None) == field for f in self._possible_fields):
                self._fields.add(field)
            else:
                raise ESMException("Illegal value for the value "+field+". The field must be in "+ str([f['name'] for f in self._possible_fields]))
        except GeneratorExit :
            pass
        except:raise

    @time_range.setter
    def time_range(self, time_range):
        """
        Set the time range of the query to the specified string value.
        """
        if time_range is not 'CUSTOM' and self._compute_time_range :
            try :
                times = getTimes(time_range)
                self._time_range='CUSTOM'
                self._start_time=times[0]
                self._end_time=times[1]

            except ESMException :
                self._session._logger.warning('The choosen time range is not fully supported. This mean that only the first page of events will be returned. '+str(ESMException))
                self._time_range=time_range
        else :
            self._time_range=time_range
  

    def execute(self):
        """"
            Execute the query.
            Returns a list of Events.
        """

        if self.time_range is 'CUSTOM' :
            self._query=self.esmRequest(
                'event_query_custom_time',
                asynch=False,

                time_range=self.time_range,
                start_time=self.start_time,
                end_time=self.end_time,
                #order=self.order, TODO support order
                fields=self.fields,
                filters=self.filters,
                limit=self.limit,
                offset=self.offset,
                includeTotal=False
                )

        else :
            self._query=self.esmRequest(
                'event_query',
                asynch=False,

                time_range=self.time_range,
                #order=self.order, TODO support order
                fields=self.fields,
                filters=self.filters,
                limit=self.limit,
                offset=self.offset,
                includeTotal=False
                )
        
        self._session._logger.debug("EsmRunningQuery object : "+str(self._query))

        if self._waitFor(self._query['resultID']):
            self._executed = True

        return (self.getEvents())

    def _waitFor(self, resultID):
        self._session._logger.debug("Waiting for the query to be executed on the SIEM...")
        while True:
            status = self.esmRequest('query_status', resultID=resultID)
            if status : 
                if status['complete'] is True :
                    return True
                else :
                    time.sleep(0.5)
            else :
                raise ESMException("There was an error while waiting for the query to execute.")

    def getEvents(self, startPos=0, numRows=None):
        if self._executed :
            if not numRows :
                numRows=self.limit
                
            result=self.esmRequest('query_result', startPos=startPos, numRows=numRows, resultID=self._query['resultID'])
            events=self._parseResult(result['columns'], result['rows'])
            return events
        else :
            raise ESMException("Query need to be executed before to be able to get events.")

    def _parseResult(self, columns, rows):
        events=list()
        for row in rows :
            event=dict()
            for i in range(len(columns)-1):
                event.update({columns[i]['name']:row['values'][i]})
            events.append(event)
        return events

class QueryFilter(ESMObject):

    _possible_filters = []

    def __init__(self):
        super().__init__()

        #Mapping singleton attributes
        self.__possible_filters = QueryFilter._possible_filters

        #Setting up static constant
        """ Not checking dynamically the velidity of the fields cause makes too much of unecessary requests
        if self._possible_filters is None :
            self._possible_filters = self._get_possible_filters()
            """

    def _get_possible_filters(self):
        return(self.esmRequest('get_possible_filters'))

    @abstractmethod
    def configDict(self):
        raise ESMException("Not implemented in the base filter")

class GroupFilter(QueryFilter):
    """
        Based on EsmFilterGroup
    """

    def __init__(self, *filters, logic='AND') :
        super().__init__()
        
        #Declaring attributes
        self._filters=filters
        self._logic=logic

    def configDict(self):
        return({
            "type": "EsmFilterGroup",
            "filters": [f.configDict() for f in self._filters],
            "logic":self._logic
            })
        
class FieldFilter(QueryFilter):
    """
    Based on EsmFieldFilter
    """

    def __init__(self, name, values, operator='IN') :
        super().__init__()
        #Declaring attributes
        self._name=str()
        self._operator=str()
        self._values=list()

        self._name = name
        self.operator = operator
        self.values = values

    def configDict(self):
        return ({
            "type": "EsmFieldFilter",
            "field": {"name": self._name},
            "operator": self._operator,
            "values": self._values
            })

    @property
    def field_name(self):
        return (self._field_name)
    
    @property
    def operator(self):
        return (self._operator)

    @property
    def values(self):
        return (self._values)

    @field_name.setter
    def field_name(self, name):
        try:
            if True : # Not checking dynamically the velidity of the fields cause makes too much of unecessary requests any(f.get('name', None) == name for f in self._possible_filters):
                self._field_name = name
            else:
                raise ESMException("Illegal value for the "+name+" field. The filter must be in :"+str([f['name'] for f in self._possible_filters]))
        except GeneratorExit:
            pass
        except:
            raise

    @operator.setter
    def operator(self, operator):
        try:
            if operator in POSSIBLE_OPERATORS :
                self._operator = operator
            else:
                raise ESMException("Illegal value for the filter operator "+operator+". The operator must be in "+str(POSSIBLE_OPERATORS))
        except:
            raise
        
    def add_value(self, type, **args):
        """
        Please refer to the EsmFilterValue documentation
        """
        try:
            type_template=None
            for possible_value_type in POSSIBLE_VALUE_TYPES :
                if possible_value_type['type'] == type :
                    type_template=possible_value_type
                    break

            if type_template is not None :
                if type_template['key'] in args :
                    self._values.append({'type':type, type_template['key']:args[type_template['key']]})

                else:
                    raise ESMException("You must provide a valid named parameter containing the value(s). The key must be in "+str(POSSIBLE_VALUE_TYPES))

            else:
                raise ESMException("Illegal value type for the value filter. The type must be in "+str(POSSIBLE_VALUE_TYPES))
        except:
            raise

    def add_basic_value(self, value):
        self.add_value(type='EsmBasicValue', value=str(value))

    @values.setter
    def values(self, values):
        for v in values :
            if isinstance(v, dict):
                self.add_value(**v)

            elif isinstance(v, (int, float, str)) :
                self.add_basic_value(v)






# -*- coding: utf-8 -*-
"""
mfesaw2 exceptions
"""

class ESMException(Exception):
    """Base Exception"""
    pass

class ESMDataSourceNotFound(ESMException):
    """Raised when the ESM returns an error while: 
    'deserializing ESMDataSourceDetail"""
    pass

class ESMParamsError(ESMException):
    pass

class ESMAuthError(ESMException):
    pass